import re
import time

from ats_base.common import func
from ats_base.log.logger import logger
from ats_base.service import app, pro, acd

from ats_case.case.context import Context
from ats_case.common.enum import ProClazz, OperationClazz
from ats_case.common.error import APIError

"""
    常用操作命令
"""


def send(context: Context, todo: dict, types=2, retry_times: int = 3):
    """
    发送操作命令 - 向测试端app
    :param context:         上下文
    :param todo:            任务
    :param types:
    :param retry_times:     失败重试次数（默认：3次）
    :return:
    """
    result = None

    try:
        data = {
            'type': types,
            'exec_time': func.sys_current_time(),
            'test_sn': context.test_sn,
            'case_id': context.case.id,
            'meter_pos': context.meter.pos,
            'step_id': context.runtime.step,
            'todo': todo
        }

        logger.info('~ @TCC-SEND-> client:{} data:{}'.format(context.tester.api, data))
        result = app.send(context.tester.api, data)
        logger.info('~ @TCC-SEND<- result:{}'.format(result))
    # except requests.exceptions.MissingSchema as me:
    #     logger.error(str(me))
    #     raise AssertionError(str(me))
    except Exception as ae:
        logger.error(str(ae))

        retry_times -= 1
        if retry_times <= 0:
            raise APIError(context.tester.api)
        else:
            sleep(5)
            send(context, todo, types, retry_times)

    return result


def sleep(seconds: float):
    """
    休眠
    :param seconds:     秒
    :return:
    """
    logger.info('~ @TCC-SLEEP-> {}secs'.format(seconds))
    time.sleep(seconds)


def offbench(context: Context, disabled=1):
    """
    脱表台
    :param context:
    :param disabled:     秒
    :return:
    """
    clazz = OperationClazz(context.case.steps[str(context.runtime.step)].get('type'))

    if disabled == 1:
        if clazz == OperationClazz.BENCH:
            return True

    return False


def cache(context: Context, data):
    """
    缓存
    :param context:
    :param data:
    :return:
    """
    pass


"""
    通讯协议篇
"""


def meter(protocol: str):
    return Meter(protocol)


class Meter(object):
    def __init__(self, protocol):
        self._protocol = ProClazz(protocol)
        self._comm_addr = None
        self._operation = None
        self._element = None
        self._parameter = None
        self._frame = None
        self._func = None
        self._func_module = None
        self._func_parameter = {}

    def comm_addr(self, addr: str):
        self._comm_addr = addr
        return self

    def operation(self, key: str):
        self._operation = key
        return self

    def element(self, di: str):
        self._element = di
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def frame(self, hexStr: str):
        self._frame = hexStr
        return self

    def encode(self):
        logger.info(
            '~ @PRO-ENCODE-> protocol:{} comm_addr:{} operation:{} element:{}'.format(self._protocol,
                                                                                      self._comm_addr,
                                                                                      self._operation,
                                                                                      self._element))
        parse = pro.encode(func.to_dict(protocol=self._protocol.name, comm_addr=self._comm_addr,
                                        operation=self._operation, element=self._element, parameter=self._parameter))
        logger.info('~ @PRO-ENCODE<- protocol:{} frame:{}'.format(self._protocol, parse.get('frame')))

        return parse.get('frame')

    def decode(self):
        logger.info('~ @PRO-DECODE-> protocol:{} frame:{}'.format(self._protocol, self._frame))
        data = pro.decode(func.to_dict(protocol=self._protocol.name, frame=self._frame))
        logger.info('~ @PRO-DECODE<- protocol:{} parse:{}'.format(self._protocol, data))

        return data.get('parse').get('link_data').get('mission').get('result')

    def function(self, data):
        self._func = data.get('code')
        self._func_module = data.get('module')
        self._func_parameter = data.get('parameter', {})

        return self

    def acv(self):
        logger.info('~ @ACD-HANDLE-> module:{} function:{} parameter:{}'.format(
            self._func_module, self._func, self._func_parameter))
        result = acd.handle(module=self._func_module, function=self._func, data=self._func_parameter)
        logger.info('~ @ACD-HANDLE<- module:{} function:{} result:{}'.format(self._func_module, self._func, result))

        return result

    def exec(self, context: Context):
        self._frame = self.encode()

        result = send(context,
                      todo={'meter:comm': {'channel': {'type': 'RS485', 'baudrate': 9600}, 'frame': self._frame}})

        self._frame = result.get('result')

        self._func_parameter['return_result'] = self.decode()

        fps = str(self._func_parameter)
        re_list = re.findall(r"#(.+?)\'", fps)
        for r in re_list:
            v = eval(r)
            fps = fps.replace('#{}'.format(r), v)

        self._func_parameter = eval(fps)

        send(context, todo={'app:show': {'msg': self.acv()}})


"""
    加密机篇
"""


def encrypt(clazz: str):
    pass


class Encryptor(object):
    def __init__(self, clazz):
        self._clazz = clazz
        pass

    def operation(self, key: str):
        pass

    def parameter(self, **param):
        pass

    def exec(self):
        pass


"""
    表台篇
"""


def bench(manufacture: str):
    pass


class Bench(object):
    def __init__(self, manufacture):
        self._manufacture = manufacture
        self._operation = None
        self._parameter = None
        self._command = None
        self._result = None

    def operation(self, command: str):
        self._operation = command
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def encode(self):
        logger.info(
            '~ @BENCH-> manufacture:{} operation:{} parameter:{}'.format(self._manufacture, self._operation,
                                                                         self._parameter))
        self._command = {'{}'.format(self._operation): self._parameter}

        return self._command

    def decode(self):
        logger.info('~ @BENCH<- manufacture:{} operation:{} result:{}'.format(self._manufacture,
                                                                              self._operation, self._result))

        return self._result

    def exec(self, context: Context):
        self.encode()
        result = send(context, todo=self._command)
        self._result = result.get('result')

        return self.decode()


"""
    测试终端篇
"""


def client():
    return App()


class App(object):
    def __init__(self):
        self._name = 'app'
        self._operation = None
        self._message = None
        self._parameter = None
        self._command = None

    def operation(self, command: str):
        self._operation = command
        return self

    def message(self, msg: str):
        self._message = msg
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def encode(self):
        logger.info('~ @APP-> operation:{} message:{}'.format(self._operation, self._message))
        self._command = {'{}:{}'.format(self._name, self._operation): self._message}

        return self._command

    def decode(self):
        pass

    def exec(self, context: Context):
        self.encode()
        send(context, todo=self._command)


"""
    平台篇
"""


def ats():
    return ATS()


class ATS(object):
    def __init__(self):
        self._name = 'ats'
        self._operation = None
        self._parameter = None

    def operation(self, command: str):
        self._operation = command
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def exec(self, context: Context):
        logger.info('~ @ATS-> operation:{} parameter:{}'.format(self._operation, self._parameter))
        return eval('{}(context, {})'.format(self._operation, self._parameter))



from importlib import import_module

from ats_base.log.logger import logger

from ats_case.case import translator, command
from ats_case.case.context import Context
from ats_case.common.enum import *


def execute(context: Context):
    if context.mode == WorkMode.FORMAL:
        FormalExecutor(context).exec()
    else:
        DebugExecutor(context).exec()


class Executor(object):
    def __init__(self, context: Context):
        self._context = context
        self._model = None

    def exec(self):
        pass

    def is_exec(self):
        ifs = self._context.case.control.get('ifs')
        if ifs is not None and type(ifs) is dict and len(ifs) > 0:
            for func, values in ifs.items():
                if command.ats().operation(func).parameter(values).exec(self._context):
                    return False
        return True

    def step_exec(self):
        try:
            logger.info('~ @TCC-STEP-> steps[#{}] execute'.format(self._context.runtime.step))

            if self.is_exec():
                result = getattr(self._model, 'step_{}'.format(self._context.runtime.step))(self._context)
                if result is not None:
                    self._context.runtime.last_step_result = result
        except Exception as e:
            logger.error(str(e))

    def loop_meet(self):
        try:
            loops = self._context.case.control.get('loops')

            if loops is None or type(loops) is not list or len(loops) <= 0:
                return False

            if self._context.runtime.loop_sn >= len(loops):
                return False
            loop = loops[self._context.runtime.loop_sn]
            ranges = loop.get('range')
            count = loop.get('count')

            step_start = int(ranges.split(':')[0])
            step_end = int(ranges.split(':')[1])

            if step_start <= self._context.runtime.step <= step_end:
                self._context.runtime.loop_start_step = step_start
                self._context.runtime.loop_end_step = step_end
                self._context.runtime.loop_count = count
                self._context.runtime.loop_index = 0
        except Exception as e:
            logger.error(str(e))
            return False

        return True

    def loop_exec(self):
        logger.info('~ @TCC-LOOP-> loops[#{}] start. -range {}:{}  -count {}'.format(
            self._context.runtime.loop_sn, self._context.runtime.loop_start_step,
            self._context.runtime.loop_end_step, self._context.runtime.loop_count))

        command.client().operation('show').message('循环[#{}]开始, 步骤范围[{}-{}], 共{}次'.format(
            self._context.runtime.loop_sn, self._context.runtime.loop_start_step,
            self._context.runtime.loop_end_step, self._context.runtime.loop_count)).exec(self._context)

        # while self._context.runtime.loop_index < self._context.runtime.loop_count:
        for i in range(1, self._context.runtime.loop_count + 1):
            self._context.runtime.loop_index = i

            logger.info('~ @TCC-LOOP-> loops[#{}], -count {}, -index {}'.format(
                self._context.runtime.loop_sn, self._context.runtime.loop_count,
                self._context.runtime.loop_index))
            command.client().operation('show').message('循环[#{}], 共{}次, 当前第{}次'.format(
                self._context.runtime.loop_sn, self._context.runtime.loop_count,
                self._context.runtime.loop_index)).exec(self._context)

            for j in range(self._context.runtime.loop_start_step, self._context.runtime.loop_end_step + 1):
                self._context.runtime.step = j
                self.step_exec()

        command.client().operation('show').message("循环结束...").exec(self._context)
        logger.info('~ @TCC-LOOP-> loops[#{}] end.'.format(self._context.runtime.loop_sn))
        self._context.runtime.loop_sn += 1


class FormalExecutor(Executor):
    def exec(self):
        # 分为两种情况: 1. 手动编写脚本 2.自动编写脚本
        pass


class DebugExecutor(Executor):
    def exec(self):
        steps = translator.translate(self._context)
        self._model = import_module('script.debug.{}.steps'.format(self._context.tester.username))

        index = 0
        while index < len(steps):
            self._context.runtime.step = steps[index]
            if self.loop_meet():
                self.loop_exec()
            else:
                self.step_exec()

            if self._context.runtime.step > (index + 1):
                index = self._context.runtime.step
            else:
                index += 1



from ats_base.service import mm


class Context(object):
    """
    测试用例上下文
    """

    def __init__(self, sn, cid, pos):
        self._test_sn = sn
        self._case_id = cid
        self._pos = pos
        self._script = None
        self._mode = None
        self._parse()

    def _parse(self):
        tl = mm.Dict.get("debug:log", self._test_sn)

        self._tester = self.Tester(tl.get('tester'))
        self._case = self.Case(tl.get('usercase'))
        self._meter = self.Meter(tl.get('meter'))
        self._bench = self.Bench(tl.get('bench'))
        self._runtime = self.Runtime()

    #     self._bench = self.Bench(tl.get('bench'))
    #     self._case = self.Case(tl.get('cases').get(self.case_id))
    #     self._meter = tl.get('meter')
    #     self._runtime = self.Runtime()
    #
    #     self._exec_result = None
    #     if self.rerun == 1:
    #         # self._exec_result = tl.get('exec_result').get(str(self._case_id))
    #         # try:
    #         #     self.runtime.loop_index = self.exec_result.get('loop_index', 0)
    #         # except:
    #         try:
    #             self.runtime.loop_index = int(self._cache.get('loop_index'))
    #         except:
    #             self.runtime.loop_index = 0

    @property
    def test_sn(self):
        return self._test_sn

    @property
    def tester(self):
        return self._tester

    @property
    def bench(self):
        return self._bench

    @property
    def case(self):
        return self._case

    @property
    def meter(self):
        return self._meter

    @property
    def mode(self):
        return self._mode

    @property
    def runtime(self):
        return self._runtime

    class Tester(object):
        def __init__(self, data: dict):
            self._ip = data.get('ip')
            self._port = data.get('port')
            self._username = data.get('username')
            self._hostname = data.get('hostname')

        @property
        def ip(self):
            return self._ip

        @property
        def port(self):
            return self._port

        @property
        def username(self):
            return self._username

        @property
        def hostname(self):
            return self._hostname

        @property
        def api(self):
            return "http://{}:{}/accept".format(self.ip, self.port)

    class Bench(object):
        def __init__(self, data: dict):
            if data is not None and len(data) > 0:
                self._manufacture = data.get('manufacture', '')
                self._model = data.get('model', '')
                self._port = data.get('port', 0)
    #         self._iabc = data.get('iabc', 'H')
    #
        @property
        def manufacture(self):
            return self._manufacture

        @property
        def model(self):
            return self._model

        @property
        def port(self):
            return self._port
    #
    #     @property
    #     def iabc(self):
    #         return self._iabc
    #
    #     @iabc.setter
    #     def iabc(self, value):
    #         self._iabc = value

    class Case(object):
        def __init__(self, data: dict):
            self._id = data.get('id', -1)
            self._name = data.get('name', 'test')
    #         self._code = data.get('code')
    #         self._version = data.get('version')
    #         self._scope = data.get('scope')
    #         self._purpose = data.get('purpose')
            self._control = data.get('control', {})
            self._steps = data.get('steps', {})
    #         self._script = 'script.{}'.format(data.get('script'))
    #         self._bench_step_orders = data.get('bench_step_orders', [])
    #         # 用户发起的参数
    #         self._req_params = self.ReqParams(data.get('req_params'))
    #
    #         lps = data.get('loops')
    #         if lps is not None and lps != '':
    #             self._loops = []
    #             lps = eval(data.get('loops'))
    #             if type(lps) is list:
    #                 for lp in lps:
    #                     self._loops.append(self.Loop(lp))
    #
    #             if self._script is None:
    #                 raise Exception('Path of script file is null.')
    #             pfs = self._script.split('.')
    #             pfs[-1] = '{}.py'.format(pfs[-1])
    #             if not util.exist(*pfs):
    #                 raise Exception('Script file[{}] is not exist.'.format(self._script))

        @property
        def id(self):
            return self._id

        @property
        def name(self):
            return self._name

        @property
        def control(self):
            return self._control

        @property
        def steps(self):
            return self._steps

    #     @property
    #     def code(self):
    #         return self._code
    #
    #     @property
    #     def version(self):
    #         return self._version
    #
    #     @property
    #     def scope(self):
    #         return self._scope
    #
    #     @property
    #     def purpose(self):
    #         return self._purpose
    #
    #     @property
    #     def loops(self):
    #         return self._loops
    #
    #     @property
    #     def script(self):
    #         return self._script
    #
    #     @property
    #     def bench_step_orders(self):
    #         return self._bench_step_orders
    #
    #     @property
    #     def req_params(self):
    #         return self._req_params

    class Meter(object):
        def __init__(self, data: dict):
            self._pos = data.get('pos')
            self._addr = data.get('addr')
            self._no = data.get('no')
    #         self._protocol = data.get('protocol')
    #         self._model = data.get('model')
    #         self._connect = data.get('connect')
    #         self._channel = {'RS485': data.get('baudrate_485'), 'IR': data.get('baudrate_ir'),
    #                          'PLC': data.get('baudrate_plc')}
    #
    #     def get_channel(self, c: CHANNEL):
    #         return func.to_dict(type=c.value, baudrate=self._channel.get(c.value))

        @property
        def pos(self):
            return self._pos

        @property
        def addr(self):
            return self._addr

        @property
        def no(self):
            return self._no

    #     @property
    #     def protocol(self):
    #         return self._protocol
    #
    #     @property
    #     def model(self):
    #         return self._model
    #
    #     @property
    #     def connect(self):
    #         return self._connect
    #
    #     @property
    #     def channel(self):
    #         return self._channel
    #
    class Runtime(object):
        """
        运行时
        """

        def __init__(self):
            self._step = -1
            self._last_step_result = None
            self._loop_start_step = 0
            self._loop_end_step = 0
            self._loop_count = 0
            self._loop_index = 0
            self._loop_sn = 0

        @property
        def step(self):
            return self._step

        @step.setter
        def step(self, value):
            self._step = value

        @property
        def last_step_result(self):
            return self._last_step_result

        @last_step_result.setter
        def last_step_result(self, value):
            self._last_step_result = value

        @property
        def loop_start_step(self):
            return self._loop_start_step

        @loop_start_step.setter
        def loop_start_step(self, value):
            self._loop_start_step = value

        @property
        def loop_end_step(self):
            return self._loop_end_step

        @loop_end_step.setter
        def loop_end_step(self, value):
            self._loop_end_step = value

        @property
        def loop_count(self):
            return self._loop_count

        @loop_count.setter
        def loop_count(self, value):
            self._loop_count = value

        @property
        def loop_index(self):
            return self._loop_index

        @loop_index.setter
        def loop_index(self, value):
            self._loop_index = value

        @property
        def loop_sn(self):
            return self._loop_sn

        @loop_sn.setter
        def loop_sn(self, value):
            self._loop_sn = value





import os
import pytest

from pathlib import Path

from datetime import datetime
from ats_base.service import db, mm

from ats_case.common.enum import WorkMode


def run(**kwargs):
    try:
        mode = WorkMode(kwargs.get('mode'))
        if mode == WorkMode.FORMAL:
            pt = FormalMode(kwargs)
        else:
            pt = DebugMode(kwargs)
        pt.run()
    except:
        pass


class ExecMode(object):
    def __init__(self, data: dict):
        self._data = data
        self._username = self._data.get('tester').get('username', '')
        self._sn = self.gen_sn()

    def run(self):
        pass

    def gen_sn(self):
        return self._now() + self._username.upper()

    def _now(self):
        return datetime.now().strftime('%y%m%d%H%M%S%f')

    def _save(self):
        pass

    def _build(self, work_mode: WorkMode, code: str = None):
        if code is None:
            code = 'case'

        root_dir = os.path.dirname(os.path.dirname(__file__))

        user_dir = os.path.join(root_dir, 'testcase', work_mode.value.lower(), self._username)
        template_file = os.path.join(root_dir, 'template', 'testcase_v1.tmp')
        script_file = os.path.join(user_dir, 'test_{}.py'.format(code))

        if not os.path.exists(user_dir):
            Path(user_dir).mkdir(parents=True, exist_ok=True)

        with open(template_file, 'r', encoding='UTF-8') as file:
            content = file.read()
            content = content.replace('{script}', code.upper())
        with open(script_file, 'w', encoding='UTF-8') as file:
            file.write(content)

        return script_file


class FormalMode(ExecMode):
    def run(self):
        self._save()

        cases = self._data.get('cases')
        meters = self._data.get('meters')

        for cid, case in cases.items():
            for i in range(len(meters)):
                # 客户端出来的参数需要缓存case, meter，context从数据库里去原始数据
                # i = 0 执行操作表台 传入参数index
                # pytest.main(
                #     ["-sv", self._build(WorkMode.FORMAL), '--sn={}'.format(self._sn), '--cid={}'.format(cid),
                #      '--rerun={}'.format(rerun)])
                pass

    def _save(self):
        mm.Dict.put('debug:log', self._sn, self._data)


class DebugMode(ExecMode):
    def __init__(self, data: dict):
        super().__init__(data)

    def run(self):
        self._save()
        pytest.main(["-sv", self._build(WorkMode.DEBUG), '--sn={}'.format(self._sn)])

    def _save(self):
        mm.Dict.put('debug:log', self._sn, self._data)

